#! /usr/bin/env sh

# defaults
[ -n "${CIDD_DNS}" ] || CIDD_DNS="\
9.9.9.9 \
"
[ -n "${CIDD_GIT_CHILD}" ] || CIDD_GIT_CHILD="cd"
[ -n "${CIDD_GIT_PARENT}" ] || CIDD_GIT_PARENT="rwx"

# main
cd_main () {
    cd_set_environment_variables
    cd_set_packages_repositories
    cd_set_packages_configuration
    #
    cd_list_working_directory
    cd_set_https_verification_off
    cd_set_dns_resolving
    cd_update_packages_catalog
    cd_install_packages_tools
    cd_install_ca_certificates
    cd_write_ca_certificates
    cd_update_ca_certificates
    cd_set_https_verification_on
    cd_update_packages_catalog
    cd_upgrade_packages
    cd_install_git
    cd_install_python
    # TODO move to Python
    cd_install_rsync
    # TODO move to Python
    cd_install_ssh
    cd_clean_packages_cache
    cd_install_python_modules
    cd_write_python_module
    cd_switch_to_python "${@}"
}

# steps

cd_set_environment_variables () {
    cd_step "Set environment variables"
    # set path
    CIDD_PATH="$(realpath "${0}")"
    cd_echo "CIDD_PATH"
    # set operating system id
    CIDD_OS_ID="$(cd_grep_os ID)"
    case "${CIDD_OS_ID}" in
        "almalinux") CIDD_OS_ID="${CIDD_OS_ALMA}" ;;
        "alpine") CIDD_OS_ID="${CIDD_OS_ALPINE}" ;;
        "arch") CIDD_OS_ID="${CIDD_OS_ARCH}" ;;
        "debian") CIDD_OS_ID="${CIDD_OS_DEBIAN}" ;;
        "fedora") CIDD_OS_ID="${CIDD_OS_FEDORA}" ;;
        "rocky") CIDD_OS_ID="${CIDD_OS_ROCKY}" ;;
        "ubuntu") CIDD_OS_ID="${CIDD_OS_UBUNTU}" ;;
        *) cd_error_os "CIDD_OS_ID" ;;
    esac
    # set operating system version
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}"|"${CIDD_OS_ARCH}"|"${CIDD_OS_FEDORA}"|"${CIDD_OS_ROCKY}")
            CIDD_OS_VERSION=$(cd_grep_os VERSION_ID \
                | sed "s|^\([0-9]\+\)\..*|\1|")
        ;;
        "${CIDD_OS_ALPINE}")
            CIDD_OS_VERSION=$(cd_grep_os VERSION_ID \
                | sed "s|^\([0-9]\+\.[0-9]\+\)\..*|\1|")
        ;;
        "${CIDD_OS_DEBIAN}"|"${CIDD_OS_UBUNTU}")
            CIDD_OS_VERSION="$(cd_grep_os VERSION_CODENAME)"
        ;;
        *)
    esac
    # check operating system version
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}"|"${CIDD_OS_ROCKY}")
            case "${CIDD_OS_VERSION}" in
                "8"|"9") ;;
                *) cd_error_os "CIDD_OS_VERSION" ;;
            esac
        ;;
        "${CIDD_OS_ALPINE}")
            case "${CIDD_OS_VERSION}" in
                "3.18"|"3.19") ;;
                *) cd_error_os "CIDD_OS_VERSION" ;;
            esac
        ;;
        "${CIDD_OS_ARCH}")
            case "${CIDD_OS_VERSION}" in
                "20231112"|"20240101") ;;
                *) cd_error_os "CIDD_OS_VERSION" ;;
            esac
        ;;
        "${CIDD_OS_DEBIAN}")
            case "${CIDD_OS_VERSION}" in
                "bookworm"|"bullseye") ;;
                *) cd_error_os "CIDD_OS_VERSION" ;;
            esac
        ;;
        "${CIDD_OS_FEDORA}")
            case "${CIDD_OS_VERSION}" in
                "39"|"40") ;;
                *) cd_error_os "CIDD_OS_VERSION" ;;
            esac
        ;;
        "${CIDD_OS_UBUNTU}")
            case "${CIDD_OS_VERSION}" in
                "jammy"|"noble") ;;
                *) cd_error_os "CIDD_OS_VERSION" ;;
            esac
        ;;
        *)
    esac
    cd_split
    cd_echo "CIDD_OS_ID" "CIDD_OS_VERSION"
    # universal
    CIDD_DNS_FILE="/etc/resolv.conf"
    CIDD_PKG_CA="ca-certificates"
    CIDD_PKG_GIT="git"
    # TODO move to Python
    CIDD_PKG_RSYNC="rsync"
    CIDD_PYTHON_ALIAS="python3"
    cd_split
    cd_echo "CIDD_DNS_FILE" "CIDD_PKG_CA" "CIDD_PKG_GIT" "CIDD_PYTHON_ALIAS"
    # set ca command & root
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}"|"${CIDD_OS_FEDORA}"|"${CIDD_OS_ROCKY}")
            CIDD_CA_ROOT="/etc/pki/ca-trust/source/anchors"
            CIDD_CMD_CA="update-ca-trust"
        ;;
        "${CIDD_OS_ALPINE}")
            CIDD_CA_ROOT="/usr/local/share/ca-certificates"
            CIDD_CMD_CA="update-ca-certificates"
        ;;
        "${CIDD_OS_ARCH}")
            CIDD_CA_ROOT="/etc/ca-certificates/trust-source/anchors"
            CIDD_CMD_CA="update-ca-trust"
        ;;
        "${CIDD_OS_DEBIAN}"|"${CIDD_OS_UBUNTU}")
            CIDD_CA_ROOT="/usr/local/share/ca-certificates"
            CIDD_CMD_CA="update-ca-certificates"
        ;;
        *)
    esac
    cd_split
    cd_echo "CIDD_CA_ROOT" "CIDD_CMD_CA"
    # set package manager
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALPINE}")
            CIDD_PM="${CIDD_PM_APK}"
        ;;
        "${CIDD_OS_DEBIAN}"|"${CIDD_OS_UBUNTU}")
            CIDD_PM="${CIDD_PM_APT}"
        ;;
        "${CIDD_OS_ALMA}"|"${CIDD_OS_FEDORA}"|"${CIDD_OS_ROCKY}")
            CIDD_PM="${CIDD_PM_DNF}"
        ;;
        "${CIDD_OS_ARCH}")
            CIDD_PM="${CIDD_PM_PACMAN}"
        ;;
        *)
    esac
    cd_split
    cd_echo "CIDD_PM"
    case "${CIDD_PM}" in
        "${CIDD_PM_DNF}")
            CIDD_PM_CLEAN="dnf clean all"
            CIDD_PM_INSTALL="dnf install --assumeyes"
            CIDD_PM_QUERY="rpm --query"
            CIDD_PM_UPDATE="dnf makecache"
            CIDD_PM_UPGRADE="dnf upgrade --assumeyes"
            CIDD_PKG_PKG=""
            CIDD_PM_CONF_PATH="/etc/dnf/dnf.conf"
            CIDD_PM_CONF_TEXT="\
[main]
best=True
clean_requirements_on_remove=True
gpgcheck=1
installonly_limit=3
skip_if_unavailable=False
"
            CIDD_PM_HTTPS_PATH="/etc/dnf/dnf.conf.d/https.conf"
            CIDD_PM_HTTPS_TEXT="\
sslverify=False
"
        ;;
        "${CIDD_PM_APK}")
            CIDD_PM_CLEAN="apk cache purge"
            CIDD_PM_INSTALL="apk add"
            CIDD_PM_QUERY="apk info"
            CIDD_PM_UPDATE="apk update"
            CIDD_PM_UPGRADE="apk upgrade"
            CIDD_PKG_PKG=""
            CIDD_PM_CONF_PATH=""
            CIDD_PM_CONF_TEXT=""
            CIDD_PM_HTTPS_PATH="/etc/apk/repositories.d/https"
            CIDD_PM_HTTPS_TEXT="\
--no-verify
"
        ;;
        "${CIDD_PM_PACMAN}")
            CIDD_PM_CLEAN="pacman --sync --clean --noconfirm"
            CIDD_PM_INSTALL="pacman --sync --noconfirm"
            CIDD_PM_QUERY="pacman --query"
            CIDD_PM_UPDATE="pacman --sync --refresh"
            CIDD_PM_UPGRADE="pacman --sync --sysupgrade --noconfirm"
            CIDD_PKG_PKG=""
            CIDD_PM_CONF_PATH=""
            CIDD_PM_CONF_TEXT=""
            CIDD_PM_HTTPS_PATH="/etc/pacman.d/https.conf"
            CIDD_PM_HTTPS_TEXT="\
SSLVerify = No
"
        ;;
        "${CIDD_PM_APT}")
            CIDD_PM_CLEAN="apt-get clean"
            CIDD_PM_INSTALL="apt-get install --assume-yes"
            CIDD_PM_QUERY="dpkg-query --show"
            CIDD_PM_UPDATE="apt-get update"
            CIDD_PM_UPGRADE="apt-get upgrade --assume-yes"
            CIDD_PKG_PKG="apt-utils"
            CIDD_PM_CONF_PATH="/etc/apt/apt.conf.d/apt.conf"
            CIDD_PM_CONF_TEXT="\
Acquire::Check-Valid-Until True;
APT::Get::Show-Versions True;
APT::Install-Recommends False;
APT::Install-Suggests False;
Dir::Etc::SourceParts \"\";
"
            CIDD_PM_HTTPS_PATH="/etc/apt/apt.conf.d/https"
            CIDD_PM_HTTPS_TEXT="\
Acquire::https::Verify-Peer False;
"
        ;;
        *)
    esac
    cd_split
    cd_echo "CIDD_PM_CLEAN" \
        "CIDD_PM_INSTALL" "CIDD_PM_QUERY" "CIDD_PM_UPDATE" "CIDD_PM_UPGRADE"
    cd_split
    cd_echo "CIDD_PKG_PKG" "CIDD_PM_CONF_PATH" "CIDD_PM_HTTPS_PATH"
    # specific
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}")
            CIDD_URL_DEFAULT="https://repo.almalinux.org/almalinux"
        ;;
        "${CIDD_OS_ALPINE}")
            CIDD_URL_DEFAULT="https://dl-cdn.alpinelinux.org/alpine"
        ;;
        "${CIDD_OS_ARCH}")
            CIDD_URL_DEFAULT="https://geo.mirror.pkgbuild.com"
        ;;
        "${CIDD_OS_DEBIAN}")
            CIDD_URL_DEFAULT="http://deb.debian.org/debian"
        ;;
        "${CIDD_OS_FEDORA}")
            CIDD_URL_DEFAULT="http://download.example/pub/fedora/linux/releases"
        ;;
        "${CIDD_OS_ROCKY}")
            CIDD_URL_DEFAULT="http://dl.rockylinux.org/\$contentdir"
        ;;
        "${CIDD_OS_UBUNTU}")
            CIDD_URL_DEFAULT="http://archive.ubuntu.com/ubuntu"
        ;;
        *)
    esac
    CIDD_URL_CHOSEN="${CIDD_URL_DEFAULT}"
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}")
            [ -n "${CIDD_URL_ALMA}" ] && CIDD_URL_CHOSEN="${CIDD_URL_ALMA}"
        ;;
        "${CIDD_OS_ALPINE}")
            [ -n "${CIDD_URL_ALPINE}" ] && CIDD_URL_CHOSEN="${CIDD_URL_ALPINE}"
        ;;
        "${CIDD_OS_ARCH}")
            [ -n "${CIDD_URL_ARCH}" ] && CIDD_URL_CHOSEN="${CIDD_URL_ARCH}"
        ;;
        "${CIDD_OS_DEBIAN}")
            [ -n "${CIDD_URL_DEBIAN}" ] && CIDD_URL_CHOSEN="${CIDD_URL_DEBIAN}" \
            || CIDD_URL_CHOSEN="https://deb.debian.org/debian"
        ;;
        "${CIDD_OS_FEDORA}")
            [ -n "${CIDD_URL_FEDORA}" ] && CIDD_URL_CHOSEN="${CIDD_URL_FEDORA}" \
            || CIDD_URL_CHOSEN="https://rpmfind.net/linux/fedora/linux/releases"
        ;;
        "${CIDD_OS_ROCKY}")
            [ -n "${CIDD_URL_ROCKY}" ] && CIDD_URL_CHOSEN="${CIDD_URL_ROCKY}" \
            || CIDD_URL_CHOSEN="https://dl.rockylinux.org/\$contentdir"
        ;;
        "${CIDD_OS_UBUNTU}")
            [ -n "${CIDD_URL_UBUNTU}" ] && CIDD_URL_CHOSEN="${CIDD_URL_UBUNTU}" \
            || CIDD_URL_CHOSEN="https://ubuntu.mirrors.ovh.net/ubuntu"
        ;;
        *)
    esac
    cd_split
    cd_echo "CIDD_URL_DEFAULT" "CIDD_URL_CHOSEN"
    # set python command & package
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}"|"${CIDD_OS_ROCKY}")
            CIDD_PYTHON_COMMAND="python3.11"
            CIDD_PYTHON_PACKAGE="python3.11"
        ;;
        "${CIDD_OS_ALPINE}")
            CIDD_PYTHON_COMMAND="python3.11"
            CIDD_PYTHON_PACKAGE="python3"
        ;;
        "${CIDD_OS_ARCH}")
            CIDD_PYTHON_COMMAND="python3.12"
            CIDD_PYTHON_PACKAGE="python"
        ;;
        "${CIDD_OS_DEBIAN}")
            case "${CIDD_OS_VERSION}" in
                "bookworm") CIDD_PYTHON_COMMAND="python3.11" ;;
                "bullseye") CIDD_PYTHON_COMMAND="python3.9" ;;
                *)
            esac
            CIDD_PYTHON_PACKAGE="python3"
        ;;
        "${CIDD_OS_FEDORA}")
            CIDD_PYTHON_COMMAND="python3.12"
            CIDD_PYTHON_PACKAGE="python3"
        ;;
        "${CIDD_OS_UBUNTU}")
            case "${CIDD_OS_VERSION}" in
                "noble") CIDD_PYTHON_COMMAND="python3.12" ;;
                "jammy") CIDD_PYTHON_COMMAND="python3.10" ;;
                *)
            esac
            CIDD_PYTHON_PACKAGE="python3"
        ;;
        *)
    esac
    # set python packages
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}"|"${CIDD_OS_FEDORA}"|"${CIDD_OS_ROCKY}")
            CIDD_PYTHON_PACKAGES="/usr/lib64/${CIDD_PYTHON_COMMAND}/site-packages"
        ;;
        "${CIDD_OS_ALPINE}"|"${CIDD_OS_ARCH}")
            CIDD_PYTHON_PACKAGES="/usr/lib/${CIDD_PYTHON_COMMAND}/site-packages"
        ;;
        "${CIDD_OS_DEBIAN}"|"${CIDD_OS_UBUNTU}")
            CIDD_PYTHON_PACKAGES="/usr/lib/${CIDD_PYTHON_ALIAS}/dist-packages"
        ;;
        *)
    esac
    cd_split
    cd_echo "CIDD_PYTHON_COMMAND" "CIDD_PYTHON_PACKAGE" "CIDD_PYTHON_PACKAGES"
    # variables
    [ -n "${CIDD_CA_1}" ] && CIDD_CA=true
    # continuous integration platform
    if [ -n "${GITHUB_ACTIONS}" ] ; then
        # github → gitea → forgejo
        if [ -n "${GITHUB_SERVER_URL}" ] ; then
            CIDD_SERVER_URL="${GITHUB_SERVER_URL}"
        else
            cd_error_ci "GITHUB_SERVER_URL"
        fi
        if [ -n "${GITHUB_REPOSITORY}" ] ; then
            CIDD_PROJECTS_GROUP="$(dirname "${GITHUB_REPOSITORY}")"
            CIDD_PROJECT_NAME="$(basename "${GITHUB_REPOSITORY}")"
        else
            cd_error_ci "GITHUB_REPOSITORY"
        fi
        if [ -n "${GITHUB_REF_NAME}" ] ; then
            CIDD_PROJECT_BRANCH="${GITHUB_REF_NAME}"
        else
            cd_error_ci "GITHUB_REF_NAME"
        fi
    elif [ -n "${GITLAB_CI}" ] ; then
        # gitlab
        if [ -n "${CI_SERVER_URL}" ] ; then
            CIDD_SERVER_URL="${CI_SERVER_URL}"
        else
            cd_error_ci "CI_SERVER_URL"
        fi
        if [ -n "${CI_PROJECT_PATH}" ] ; then
            CIDD_PROJECTS_GROUP="$(dirname "${CI_PROJECT_PATH}")"
            CIDD_PROJECT_NAME="$(basename "${CI_PROJECT_PATH}")"
        else
            cd_error_ci "CI_PROJECT_PATH"
        fi
        if [ -n "${CI_COMMIT_BRANCH}" ] ; then
            CIDD_PROJECT_BRANCH="${CI_COMMIT_BRANCH}"
        else
            cd_error_ci "CI_COMMIT_BRANCH"
        fi
    else
        # unsupported
        cd_error_ci "ø"
    fi
    [ -n "${CIDD_SERVER_URL}" ] || cd_error_ci "CIDD_SERVER_URL"
    [ -n "${CIDD_PROJECTS_GROUP}" ] || cd_error_ci "CIDD_PROJECTS_GROUP"
    [ -n "${CIDD_PROJECT_NAME}" ] || cd_error_ci "CIDD_PROJECT_NAME"
    [ -n "${CIDD_PROJECT_BRANCH}" ] || cd_error_ci "CIDD_PROJECT_BRANCH"
    #
    CIDD_PROJECTS_URL="${CIDD_SERVER_URL}/${CIDD_PROJECTS_GROUP}"
    #
    cd_split
    cd_echo "CIDD_CA"
    cd_split
    cd_echo "CIDD_SERVER_URL" \
        "CIDD_PROJECTS_GROUP" "CIDD_PROJECT_NAME" "CIDD_PROJECT_BRANCH"
    cd_split
    cd_echo "CIDD_PROJECTS_URL"
    # TODO move to Python
    case "${CIDD_PM}" in
        "${CIDD_PM_APK}"|"${CIDD_PM_APT}") CIDD_PKG_SSH="openssh-client" ;;
        "${CIDD_PM_DNF}") CIDD_PKG_SSH="openssh-clients" ;;
        "${CIDD_PM_PACMAN}") CIDD_PKG_SSH="openssh" ;;
        *)
    esac
}

cd_set_packages_repositories () {
    cd_step "Set packages repositories"
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_ALMA}")
            case "${CIDD_OS_VERSION}" in
                "8") \
cd_set_packages_repositories__file="/etc/yum.repos.d/almalinux.repo" ;;
                "9") \
cd_set_packages_repositories__file="/etc/yum.repos.d/almalinux-baseos.repo" ;;
                *)
            esac
            cd_sed "${cd_set_packages_repositories__file}" \
            "|^mirrorlist|# mirrorlist|" \
            "|${CIDD_URL_DEFAULT}|${CIDD_URL_CHOSEN}|" \
            "|^# baseurl|baseurl|"
        ;;
        "${CIDD_OS_ALPINE}")
            cd_set_packages_repositories__file="/etc/apk/repositories"
            cd_write "${cd_set_packages_repositories__file}" "\
${CIDD_URL_CHOSEN}/v${CIDD_OS_VERSION}/main
${CIDD_URL_CHOSEN}/v${CIDD_OS_VERSION}/community
"
        ;;
        "${CIDD_OS_DEBIAN}")
            cd_set_packages_repositories__file="/etc/apt/sources.list"
            cd_write "${cd_set_packages_repositories__file}" "\
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION} main
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION}-backports main
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION}-updates main
deb ${CIDD_URL_CHOSEN}-security ${CIDD_OS_VERSION}-security main
"
        ;;
        "${CIDD_OS_ROCKY}")
            case "${CIDD_OS_VERSION}" in
                "8") \
cd_set_packages_repositories__file="/etc/yum.repos.d/Rocky-BaseOS.repo" ;;
                "9") \
cd_set_packages_repositories__file="/etc/yum.repos.d/rocky.repo" ;;
                *)
            esac
            cd_sed "${cd_set_packages_repositories__file}" \
            "|^mirrorlist|# mirrorlist|" \
            "|${CIDD_URL_DEFAULT}|${CIDD_URL_CHOSEN}|" \
            "|^#baseurl|baseurl|"
        ;;
        "${CIDD_OS_UBUNTU}")
            cd_set_packages_repositories__file="/etc/apt/sources.list"
            cd_write "${cd_set_packages_repositories__file}" "\
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION} main
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION}-backports main
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION}-updates main
deb ${CIDD_URL_CHOSEN} ${CIDD_OS_VERSION}-security main
"
        ;;
        *)
    esac
}

cd_set_packages_configuration () {
    cd_step "Set packages configuration"
    cd_write "${CIDD_PM_CONF_PATH}" "${CIDD_PM_CONF_TEXT}"
    case "${CIDD_OS_ID}" in
        "${CIDD_OS_DEBIAN}"|"${CIDD_OS_UBUNTU}")
            export DEBIAN_FRONTEND="noninteractive"
        ;;
        *)
    esac
}

# agnostic steps

cd_list_working_directory () {
    cd_step "List working directory"
    cd_list_working_directory__path="$(realpath .)"
    cd_ls "${cd_list_working_directory__path}"
}

cd_set_https_verification_off () {
    if [ -n "${CIDD_CA}" ] || [ "${CIDD_PM}" = "${CIDD_PM_APT}" ] ; then
        cd_step "Set HTTPS verification off"
        cd_mkdir "$(dirname "${CIDD_PM_HTTPS_PATH}")"
        cd_write "${CIDD_PM_HTTPS_PATH}" "${CIDD_PM_HTTPS_TEXT}"
    fi
}

cd_set_dns_resolving () {
    cd_step "Set DNS resolving"
    for cd_set_dns_resolving__server in ${CIDD_DNS} ; do
        cd_set_dns_resolving__text="${cd_set_dns_resolving__text}\
nameserver ${cd_set_dns_resolving__server}
"
    done
    cd_write "${CIDD_DNS_FILE}" "${cd_set_dns_resolving__text}"
}

cd_update_packages_catalog () {
    cd_step "Update packages catalog"
    ${CIDD_PM_UPDATE} || exit
}

cd_install_packages_tools () {
    cd_step "Install packages tools"
    cd_install_package "${CIDD_PKG_PKG}"
}

cd_install_ca_certificates () {
    cd_step "Install CA"
    cd_install_package "${CIDD_PKG_CA}"
}

cd_write_ca_certificates () {
    cd_step "Write CA certificates"
    cd_mkdir "${CIDD_CA_ROOT}"
    cd_write_ca_certificates__index=1
    eval "cd_write_ca_certificates__text=\
\"\${CIDD_CA_${cd_write_ca_certificates__index}}\""
    while [ -n "${cd_write_ca_certificates__text}" ] ; do
        cd_write_ca_certificates__path="\
${CIDD_CA_ROOT}/${cd_write_ca_certificates__index}.crt"
        cd_split
        cd_write \
        "${cd_write_ca_certificates__path}" \
        "${cd_write_ca_certificates__text}"
        cd_openssl "${cd_write_ca_certificates__path}"
        cd_write_ca_certificates__index=$((cd_write_ca_certificates__index+1))
        eval "cd_write_ca_certificates__text=\
\"\${CIDD_CA_${cd_write_ca_certificates__index}}\""
    done
}

cd_update_ca_certificates () {
    cd_step "Update CA certificates"
    ${CIDD_CMD_CA} || exit
}

cd_set_https_verification_on () {
    cd_step "Set HTTPS verification on"
    cd_rm "${CIDD_PM_HTTPS_PATH}"
}

cd_upgrade_packages () {
    cd_step "Upgrade packages"
    ${CIDD_PM_UPGRADE} || exit
}

cd_install_git () {
    cd_step "Install Git"
    cd_install_package "${CIDD_PKG_GIT}"
}

cd_install_python () {
    cd_step "Install Python"
    cd_install_package "${CIDD_PYTHON_PACKAGE}"
    cd_split
    cd_ln_python "${CIDD_PYTHON_COMMAND}"
}

# TODO move to Python
cd_install_rsync () {
    cd_step "Install Rsync"
    cd_install_package "${CIDD_PKG_RSYNC}"
}

# TODO move to Python
cd_install_ssh () {
    cd_step "Install SSH"
    cd_install_package "${CIDD_PKG_SSH}"
}

cd_clean_packages_cache () {
    cd_step "Clean packages cache"
    ${CIDD_PM_CLEAN} || exit
}

cd_install_python_modules () {
    cd_step "Install Python modules"
    cd_install_python_modules__root="$(mktemp --directory)" || exit
    echo "→ ${cd_install_python_modules__root}"
    for cd_install_python_modules__repository \
    in "${CIDD_GIT_CHILD}" "${CIDD_GIT_PARENT}" ; do
        cd_split
        cd_install_python_modules__url="\
${CIDD_PROJECTS_URL}/${cd_install_python_modules__repository}"
        echo "\
${cd_install_python_modules__url}
↓"
        git clone \
        "${cd_install_python_modules__url}" \
        "${cd_install_python_modules__root}\
/${cd_install_python_modules__repository}" \
        || exit
        cd_install_python_modules__path="\
${cd_install_python_modules__root}\
/${cd_install_python_modules__repository}\
/${cd_install_python_modules__repository}"
        echo "\
${cd_install_python_modules__path}
↓
${CIDD_PYTHON_PACKAGES}"
        cp --recursive \
        "${cd_install_python_modules__path}" "${CIDD_PYTHON_PACKAGES}" \
        || exit
    done
    cd_split
    cd_ls "${CIDD_PYTHON_PACKAGES}"
    cd_split
    cd_rm "${cd_install_python_modules__root}"
}

cd_write_python_module () {
    cd_step "Write Python module"
    for cd_write_python_module__variable \
    in OPEN DOWN VERT SPLT __UP SHUT OS_ID OS_VERSION ; do
        cd_write_python_module__value="\
$(cd_echo "CIDD_${cd_write_python_module__variable}")"
        cd_write_python_module__text="${cd_write_python_module__text}\
${cd_write_python_module__value}
"
    done
    cd_write "${CIDD_PYTHON_PACKAGES}/env.py" "${cd_write_python_module__text}\
CIDD_STEP = $((CIDD_STEP+1))
"
}

cd_switch_to_python () {
    cd_step "Switch to Python"
    echo "\
${CIDD_PATH}
↓
${CIDD_PYTHON_PACKAGES}/${CIDD_GIT_CHILD}"
    "${CIDD_PYTHON_ALIAS}" -m "${CIDD_GIT_CHILD}" "${@}"
}

# functions

cd_cat () {
    cd_cat__file="${1}"
    if [ -n "${cd_cat__file}" ] ; then
        cd_open "${cd_cat__file}"
        cat "${cd_cat__file}" || exit
        cd_shut "${cd_cat__file}"
    fi
}

cd_echo () {
    if [ -n "${1}" ] ; then
        for cd_echo__name in "${@}" ; do
            eval "cd_echo__text=\"\${${cd_echo__name}}\""
            echo "${cd_echo__name} = \"${cd_echo__text}\""
        done
    fi
}

cd_error_ci () {
    echo "× CI: ${*}"
    exit "${CIDD_ERROR_CI}"
}

cd_error_os () {
    cd_error_os__variable="${1}"
    printf "× OS: "
    cd_echo "${cd_error_os__variable}"
    exit "${CIDD_ERROR_OS}"
}

cd_grep_os () {
    cd_grep_os__variable="${1}"
    if [ -n "${cd_grep_os__variable}" ] ; then
        grep "^${cd_grep_os__variable}=" "/etc/os-release" \
        | sed "s|^${cd_grep_os__variable}=||" \
        | sed "s|^\"\(.*\)\"$|\1|"
    fi
}

cd_install_package () {
    cd_install_package__name="${1}"
    if [ -n "${cd_install_package__name}" ] ; then
        ${CIDD_PM_INSTALL} "${cd_install_package__name}" || exit
    fi
}

cd_ln_python () {
    cd_ln_python__command="${1}"
    if [ -n "${cd_ln_python__command}" ] ; then
        echo "→ ${CIDD_PYTHON_ALIAS} → ${cd_ln_python__command}"
        ln -f -s "${cd_ln_python__command}" "/usr/bin/${CIDD_PYTHON_ALIAS}" \
        || exit
    fi
}

cd_ls () {
    cd_ls__path="${1}"
    if [ -n "${cd_ls__path}" ] ; then
        cd_open "${cd_ls__path}"
        ls -a -l "${cd_ls__path}" || exit
        cd_shut "${cd_ls__path}"
    fi
}

cd_mkdir () {
    cd_mkdir__path="${1}"
    if [ -n "${cd_mkdir__path}" ] ; then
        echo "→ ${cd_mkdir__path}"
        mkdir --parents "${cd_mkdir__path}" || exit
    fi
}

cd_open () {
    echo "${CIDD_OPEN}${*}"
}

cd_openssl () {
    cd_openssl__file="${1}"
    if [ -f "${cd_openssl__file}" ] ; then
        openssl x509 \
        -in "${cd_openssl__file}" \
        -noout -text \
        || exit
    fi
}

cd_rm () {
    cd_rm__path="${1}"
    if [ -e "${cd_rm__path}" ] ; then
        echo "← ${cd_rm__path}"
        rm -r "${cd_rm__path}" || exit
    fi
}

cd_sed () {
    cd_sed__file="${1}"
    shift
    if [ -f "${cd_sed__file}" ] ; then
        cd_cat "${cd_sed__file}"
        for cd_sed__regex in "${@}" ; do
            sed --in-place "s${cd_sed__regex}g" "${cd_sed__file}" \
            && cd_cat "${cd_sed__file}" \
            || exit
        done
    fi
}

cd_shut () {
    echo "${CIDD_SHUT}${*}"
}

cd_split () {
    echo "${CIDD_SPLT}"
}

cd_step () {
    CIDD_STEP=$((CIDD_STEP+1))
    echo "\
${CIDD_DOWN}
${CIDD_VERT} ${CIDD_STEP} ${*}
${CIDD___UP}"
}

cd_write () {
    cd_write__file="${1}"
    cd_write__text="${2}"
    if [ -n "${cd_write__file}" ] ; then
        [ -f "${cd_write__file}" ] && cd_cat "${cd_write__file}"
        echo "→ ${cd_write__file}"
        printf "%s" "${cd_write__text}" > "${cd_write__file}" || exit
        cd_cat "${cd_write__file}"
    fi
}

# constants

CIDD_BOX_DOWN="╭"
CIDD_BOX_LEFT="╴"
CIDD_BOX_RIGHT="╶"
CIDD_BOX_UP="╰"
CIDD_BOX_VERTICAL="│"

CIDD_ERROR_CI=2
CIDD_ERROR_OS=1

CIDD_OS_ALMA="alma"
CIDD_OS_ALPINE="alpine"
CIDD_OS_ARCH="arch"
CIDD_OS_DEBIAN="debian"
CIDD_OS_FEDORA="fedora"
CIDD_OS_ROCKY="rocky"
CIDD_OS_UBUNTU="ubuntu"

CIDD_PM_APK="apk"
CIDD_PM_APT="apt"
CIDD_PM_DNF="dnf"
CIDD_PM_PACMAN="pacman"

CIDD_HORIZONTAL="────╌╌╌╌┄┄┄┄┈┈┈┈"

CIDD_OPEN="${CIDD_BOX_DOWN}${CIDD_BOX_LEFT}"
CIDD_DOWN="${CIDD_BOX_DOWN}${CIDD_HORIZONTAL}"
CIDD_VERT="${CIDD_BOX_VERTICAL}"
CIDD_SPLT="${CIDD_BOX_RIGHT}${CIDD_HORIZONTAL}"
CIDD___UP="${CIDD_BOX_UP}${CIDD_HORIZONTAL}"
CIDD_SHUT="${CIDD_BOX_UP}${CIDD_BOX_LEFT}"

# run
cd_main "${@}"
